/***************************************************************************
 *   Copyright (C) 2003, 2004 by Thomas Nagy                               *
 *   tnagy256@yahoo.fr                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <assert.h>

#include <qprocess.h>

#include <kapplication.h>
#include <kdebug.h>
#include <kstandarddirs.h>

#include <unistd.h>
#include <sys/time.h>

#include "computationdata.h"
#include "packingdata.h"
#include "algoproc.h"
#include "settings.h"


algoproc::algoproc(const QString &progname, int idx) : QObject(), algobase()
{
    m_progname = progname;
    m_idx = idx;

    loadSettings();
 
    m_height = 0;
}

algoproc::~algoproc() 
{

}

void algoproc::start()
{
    // DEBUG
    // kdWarning()<<"algoproc::start with plugin : "<<name()<<endl;

    m_data->m_status = 0;
    m_data->m_computing = true;
    m_cancompute = true;
    m_height = 0;
    algobase::start();
    
    // the command line is : ./m_progname -s
    // the algorithm parameters are given in stdin (first line) ..
    QProcess *proc = new QProcess();
    
    // DEBUG
    // kdWarning()<<"** using plugin : "<<locate( "appdata", "plugins/"+m_progname )<<endl;
    
    proc->addArgument( "nice" );
    proc->addArgument( "-n" );
    proc->addArgument( QString::number( Settings::nicevalue() ) );
    
    proc->addArgument( locate( "appdata", "plugins/"+m_progname ) );
    proc->addArgument( "-s" );
    //proc->addArgument( QString::number(m_idx) );

    //kdWarning()<<"running process with arguments : "<<proc->arguments().join(" ")<<endl;
    
    proc->start();

    QString parameters;
    parameters += QString::number( m_idx );
    parameters += " ";
    parameters += QString::number( m_data->nbValidRect() );
    parameters += " ";
    parameters += QString::number( m_data->LB() );
    parameters += " ";
    parameters += QString::number( m_data->m_adata.m_iter );
    parameters += " ";
    parameters += QString::number( m_data->m_adata.m_population );
    parameters += " ";
    parameters += QString::number( m_data->m_adata.m_mutations );
    parameters += " ";
    parameters += QString::number( m_data->m_adata.m_crossovers );
    parameters += " ";
    parameters += QString::number( 2 ); //m_data->m_adata.m_fitnesstype );
    parameters += "\n";
    
    // DEBUG
    //kdWarning()<<"using these parameters : "<<parameters<<endl;
    
    proc->writeToStdin( parameters );
    for (int i=0; i<m_data->nbRect(); i++)
    {
	if (m_data->isValidRect(i))
	{
	    rectangle r = m_data->getRect(i);
	    proc->writeToStdin( QString::number(r.r_larg)+" "+QString::number(r.r_long)+"\n" );
	    //kdWarning()<<QString::number(r.r_larg)+" "+QString::number(r.r_long)<<endl;
	}
    }
  
    while ( proc->isRunning() && m_cancompute )
    {
	while (proc->canReadLineStdout())
	{
	    m_buf = proc->readLineStdout();
	    decodeBufLine();
	    //kdWarning()<<"read this from stdout :"<<m_buf<<endl;
	}
	usleep(100000);
    }

    usleep(100000);
    while (proc->canReadLineStdout())
    {
	m_buf = proc->readLineStdout();
	decodeBufLine();
	//kdWarning()<<"read this from stdout :"<<m_buf<<endl;
    }

    // DEBUG
    /*
    if (proc->isRunning())
	kdWarning()<<"the process is still running"<<endl;
    else
	kdWarning()<<"process exited with status : "<<proc->exitStatus()<<endl;
	
    kdWarning()<<"killing the process!"<<endl;
    */	
    proc->kill();

    delete proc;

    algobase::stop();
    updateAlgoResults();
    m_cancompute=false;
    m_data->m_computing=false;
}

void algoproc::decodeBufLine()
{
    //DEBUG
    //kdWarning()<<"read buf line : "<<m_buf<<endl;
    
    if ( m_buf.startsWith("#"))
    {
	QStringList stat = QStringList::split( "\t", m_buf );
	m_data->m_status = stat[2].toInt();
	return;
    }
    
    if ( m_buf.startsWith(">>") )
    {
	if (! m_buf.startsWith(">>#") )
	    m_data->m_cdata.m_textlog.append( m_buf.right(m_buf.length() - 2 )+"\n" );
	return;
    }
    
    int count = 0;
    QStringList args = QStringList::split( "\t", m_buf );
    for (int i=0; i<args.count(); i++)
    {
	QStringList rect = QStringList::split( "|", args[i] );
	rectangle r(rect[0].toInt(), rect[1].toInt(), rect[2].toInt(), rect[3].toInt(), 
		! (bool) rect[4].toInt(), rect[5].toInt());
	
	// special note for programmers :
	// make sure the rectangles read are
	// valid or (very bad) crashes may happen here
	while (! m_data->isValidRect(count) && count < 10000) // 10000, just to make sure 
	    count++;
	
	assert(count < 5000);
	
	m_data->setRect(count, r);
	count++;
    }
    updateAlgoResults();
}

void algoproc::stop()
{
    algobase::stop();
}

int algoproc::currentHeight()
{
    return m_height;
}

// is this really needed ?
void algoproc::cleanup()
{

}

QString algoproc::name()
{
    return m_name;
}

QString algoproc::description()
{
    return m_descr;
}

void algoproc::setParams()
{
    m_buf = "";
    
    // the command line is : ./m_progname -p m_idx
    QProcess *proc = new QProcess();
    
    // DEBUG
    //kdWarning()<<"searching path for progname : "<<m_progname<<endl;
    //kdWarning()<<"** using plugin : "<<locate( "appdata", "plugins/"+m_progname )<<endl;
    
    proc->addArgument( locate( "appdata", "plugins/"+m_progname ) );
    proc->addArgument( "-p" );
    proc->addArgument( QString::number(m_idx) );

    // should renice somehow
    proc->start();

    while (!proc->canReadLineStdout())
    {
        KApplication::kApplication()->processEvents();
        // wait 250 microseconds, could be even less
        usleep(250);
        
	// DEBUG
	//kdWarning()<<"algobase::loadSettings waiting .."<<endl;
    }

    m_buf = proc->readLineStdout();

    QStringList lst = QStringList::split( "\t", m_buf );

    m_name="none, you have found a bug";
    m_descr="none, you have found a bug";

    if (lst.size() >= 7)
    {
	m_name  = QString(lst[0]);
	m_descr = QString(lst[1]);

	m_defaultsettings.m_iter        = lst[2].toInt();
	m_defaultsettings.m_population  = lst[3].toInt();
	m_defaultsettings.m_mutations   = lst[4].toInt();
	m_defaultsettings.m_crossovers  = lst[5].toInt();
	m_defaultsettings.m_maxtime     = lst[6].toInt();
	m_defaultsettings.m_fitnesstype = lst[7].toInt();
    }

    delete proc;

    m_isready = true;    
}

#include "algoproc.moc"
