/***************************************************************************
 *   Copyright (C) 2003, 2004 by Thomas Nagy                               *
 *   tnagy256@yahoo.fr                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <sys/time.h>

#include <qwmatrix.h> 
#include <qfontmetrics.h>
#include <qpainter.h>
#include <qcursor.h>
#include <qtimer.h>
#include <qpixmap.h>
#include <qmutex.h>

#include <kdebug.h>
#include <kurl.h>

#include "computationdata.h"
#include "packingdata.h"
#include "canvas_rect.h"
#include "canvas_view.h"
#include "settings.h"

#include "constants.h"

    canvas_view::canvas_view(QWidget * parent, packingdata *data) :
QCanvasView(parent) 
{
    m_data = data;
    m_updating = false;
    
    canvas_rect* r = NULL;
    for (int i=0; i<m_data->nbRect(); i++)
	m_rectlist.push_back( r );

    m_selectedRectangle = -1;
    viewport()->setMouseTracking(true);

    setCanvas( new QCanvas(800,300) );
    canvas()->setAdvancePeriod(40);

    canvas_rect::s_dofish=false;

    initColors();

    setHScrollBarMode(QScrollView::AlwaysOff);
    setVScrollBarMode(QScrollView::AlwaysOff);
    showMaximized();

    update();
    
    m_timer = new QTimer( this );
    connect( m_timer, SIGNAL(timeout()), this, SLOT(check_rectangle()) );
    m_timer->start(250, false);
}

void canvas_view::setRectColor(int i)
{
    canvas_rect *rect = m_rectlist[i];
    switch (Settings::colorMode())
    {
	case Settings::EnumColorMode::RandomPlain:
	    rect->colorer(*m_couleursPlain.at(i));
	    break;
	case Settings::EnumColorMode::RandomLight:
	    rect->colorer(*m_couleursLight.at(i));
	    break;
	case Settings::EnumColorMode::GrayScale:
	    rect->colorer(*m_couleursGray.at(i));
	    break;
	default:
	    rect->colorer(QColor(255,255,255));
	    break;
    }
}

void canvas_view::updateRatio()
{
    // warning, the drawing is rotated
    float height = (float) canvas()->width();
    float width = (float) canvas()->height();

    // first try : see if it fits with the widths
    m_ratio = width*(RATIO_GAP/(RATIO_GAP+1.0))/m_data->LB();

    m_xdecalage = (width - m_ratio*m_data->LB())/2.0;
    m_ydecalage = m_xdecalage;

    if (m_ratio*m_data->Haut() + 2*m_xdecalage > height)
    {
	// second try : with the heights
	m_ratio = height*(RATIO_GAP/(RATIO_GAP+1.0))/m_data->Haut();
	m_ydecalage = height/(2.0*RATIO_GAP);
	m_xdecalage = (width - m_ratio*m_data->LB())/2.0;
    }
}

void canvas_view::clear()
{
    while (! tryLock() )
    {
	kdWarning()<<"canvas_view::resizeEvent tryLock"<<endl;
    }

    for (int i=0; i<m_data->nbRect(); i++)
    {
	if (m_rectlist[i])
	    m_rectlist[i]->hide();
	delete m_rectlist[i];
	m_rectlist[i] = NULL;
    }

    unlock();
}

void canvas_view::resizeEvent(QResizeEvent * e)
{
    while (! tryLock() )
    {
	kdWarning()<<"canvas_view::resizeEvent tryLock"<<endl;
    }

    canvas()->resize(e->size().width(), e->size().height());
    //initialize(width(), height());
    updateRatio();

    for (int i=0; i<m_data->nbRect(); i++)
    {
	if (m_rectlist[i]) updateRectPos(i);
    }

    unlock();
}

void canvas_view::leaveEvent( QEvent * )
{
    //kdWarning()<<"leave event"<<endl;
    canvas_rect::s_dofish=false;
    checkAllFish();
    repaint(true);

    emit over("");
}

void canvas_view::enterEvent( QEvent * )
{
    //kdWarning()<<"enter event"<<endl;
    while (!tryLock())
    {
	kdWarning()<<"canvas_view::enterEvent tryLock"<<endl;
    }

    canvas_rect::s_dofish=true;
    checkAllFish();
    repaint(true);

    unlock();
}

void canvas_view::mousePressEvent(QMouseEvent* /* e */)
{
    // XXX regression - the user should be able to modify 
    //     the packing by drag and drop
    /*
       QPoint p = inverseWorldMatrix().map(e->pos());
       QCanvasItemList l=canvas()->collisions(p);
       for (QCanvasItemList::Iterator it=l.begin(); it!=l.end(); ++it) {
       if ( (*it)->rtti() == canvas_rectRTTI ) {
       canvas_rect* item= (canvas_rect*)(*it);
       if (m_selectedRectangle != -1 && m_selectedRectangle != item->m_id)
       emit swap(item->m_id, m_selectedRectangle);
       else if (item->click())
       m_selectedRectangle = item->m_id;
       else
       m_selectedRectangle = -1;
       }
       m_moving = *it;
       m_moving_start = p;
       return;
       }*/
}

void canvas_view::mouseDoubleClickEvent(QMouseEvent * e)
{
    QPoint p = inverseWorldMatrix().map(e->pos());
    QCanvasItemList l=canvas()->collisions(p);
    for (QCanvasItemList::Iterator it=l.begin(); it!=l.end(); ++it) {
	if ( (*it)->rtti() == canvas_rectRTTI ) {
	    //m_id = ((canvas_rect*) (*it))->m_id;
	    //emit flip( ((canvas_rect*) (*it))->m_id );
	    m_moving = 0;
	}
	return;
    }
}

void canvas_view::contentsMouseMoveEvent(QMouseEvent *e)
{
    if (m_updating)
	return;
    
    if ( m_data->m_computing )
    {
	canvas_rect::s_dofish=false;
	checkAllFish();

	emit over("");
	return;
    }

    while (!tryLock())
    {
	kdWarning()<<"canvas_view::contentsMouseMoveEvent tryLock()"<<endl;
    }

    canvas_rect::setDefCenter( e->pos() );
    m_moving_current = e->pos();
    checkAllFish();

    m_updating=false;
}

void canvas_view::check_rectangle()
{
    while (!tryLock())
    {
	kdWarning()<<"canvas_view::check_rectangle tryLock"<<endl;
    }

    if (! isVisible() )
    {
	unlock();
	return;
    }

    /* int cur_x = m_moving_current.x();
       int cur_y = m_moving_current.y();

       for (int i=0; i< m_data->nbRect(); i++)
       {
       canvas_rect *r = m_rectlist[i];
       if (!r)
       continue;
       if ( r->x() < cur_x && cur_x < r->x() + r->real_width() )
       {
       if ( r->y() < cur_y && cur_y < r->y() + r->real_height() )
       {
    //kdWarning()<<i<<"    "<<tmprect.identifikation()<<endl;
    //kdWarning()<<cur_x<<" "<<cur_y<<endl;
    //m_rectlist[i]->m_rect.print();

    emit over(m_rectlist[i]->m_rect.identifikation());
    m_oldhover = m_rectlist[i];
    return;
    }
    }
    } */

    QCanvasItemList l=canvas()->collisions(m_moving_current);
    for (QCanvasItemList::Iterator it=l.begin(); it!=l.end(); ++it) 
    {
	if ( (*it)->rtti() == canvas_rectRTTI )
	{
	    canvas_rect* rect = (canvas_rect*) (*it);
	    if (rect != m_oldhover)
	    {
		emit over(m_data->getRect(rect->id()).identifikation());
		m_oldhover = rect;
	    }
	    unlock();
	    return;
	}
    }

    m_oldhover = NULL;
    emit over("");

    unlock();
}

void canvas_view::checkAllFish()
{
    for (int i=0; i< m_data->nbRect(); i++)
    {
	if (m_rectlist[i])
	    m_rectlist[i]->checkFish();
    }
}

void canvas_view::updateRect(int i)
{
    if (m_rectlist[i])
    {
	if (! (Settings::showOrder() || Settings::showRotation() || Settings::showIndex()) )
	{
	    m_rectlist[i]->setVisibleLabel( false );
	    return;
	}

	QString label;
       
	// show the id of the box
	if (Settings::showIndex())
	    label += QString::number(i+1);

	// show the order of the box
	if (Settings::showOrder())
	{
	    label += "(";
	    label += QString::number(m_data->getRect(i).numero);
	    label += ")";
	}
	
	// show the box rotation
	if (Settings::showRotation())
	{
	    rectangle r = m_data->getRect(i);
	    if (r.longueur() > r.largeur())
		label += "_";
	    else
		label += "|";
	}
	
	m_rectlist[i]->setLabel( label );
	m_rectlist[i]->setId( i );

	m_rectlist[i]->setVisibleLabel( true );

	updateRectPos(i);
	m_rectlist[i]->checkFish();
    }
}

void canvas_view::updateRectPos(int i)
{
    if (!m_rectlist[i])
    {
	//kdWarning()<<"BUG canvas_view::updateRectPos"<<endl;
	return;
    }

    //kdWarning()<<"canvas_view::updateRectPos"<<endl;

    setRectColor(i);
    rectangle r = m_data->getRect(i);
    m_rectlist[i]->setDims(
	    m_ratio * r.longueur(),
	    m_ratio * r.largeur(),
	    m_ydecalage + m_ratio * r.posy,
	    m_xdecalage + m_ratio * r.posx
	    );
    m_rectlist[i]->checkFish();

    // DEBUG
    //m_rectlist[i]->printSize();
}

void canvas_view::initColors()
{
    for (int i=0; i< m_data->nbRect(); i++)
    {
	m_couleursPlain += QColor(rand()%200+55, rand()%200+55, rand()%200+55);
	m_couleursLight += QColor(rand()%100+155, rand()%100+155, rand()%100+155);

	int gr = rand()%100+155;
	m_couleursGray += QColor(gr, gr, gr);
    }
}

void canvas_view::exportSVG(QTextStream* stream)
{
    int width = m_data->LB();
    int height= m_data->Haut();

    (*stream)<<"<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>"<<endl;
    (*stream)<<"<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.0//EN\""<<endl;
    (*stream)<<"\"http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd\">"<<endl;
    (*stream)<<"<svg xmlns=\"http://www.w3.org/2000/svg\" xmlns:xlink=\"http://www.w3.org/1999/xlink\" version=\"1.0\""<<endl;
    (*stream)<<"   x=\"0\" y=\"0\" width=\""<<width<<"\" height=\""<<height<<"\""<<endl;
    (*stream)<<"   id=\"svg1\" xml:space=\"preserve\">"<<endl;

    for (int i=0; i<m_data->nbRect(); i++)
    {
	int X = m_data->posX(i);
	int Y = m_data->posY(i);
	width = m_data->width(i);
	height = m_data->height(i);

	if ( !m_data->isValidRect(i) )
	    continue;

	(*stream)<<"<rect"<<endl;
	(*stream)<<"   width=\""<<width<<"\" height=\""<<height<<"\""<<endl;
	(*stream)<<"   x=\""<<X<<"\" y=\""<<Y<<"\""<<endl;
	(*stream)<<"   style=\"font-size:12;fill:#df3032;fill-opacity:0.7;fill-rule:evenodd;stroke:#000000;\""<<endl;
	(*stream)<<"/>"<<endl;
    }

    (*stream)<<"</svg>"<<endl;
}

void canvas_view::exportPNG(QString str)
{
    QPixmap pix(canvas()->width(), canvas()->height());
    QPainter p(&pix);
    canvas()->drawArea( canvas()->rect(), &p );

    pix.save( str, "PNG" );
}

void canvas_view::update()
{
    // DEBUG
    //kdWarning()<<"\033[1;33m canvas_view::update  \033[0m"<<endl;
    
    while (!tryLock())
    {
	kdWarning()<<"locking canvas_view::update"<<endl;
    }

    canvas()->setAdvancePeriod(-1);

    updateRatio();

    bool reset = false;
    if (canvas_rect::s_dofish)
    {
	canvas_rect::s_dofish=false;
	reset = true;
    }

    //kdWarning()<<"canvas_view::update"<<endl;
    for (int i=0; i<m_data->nbRect(); i++)
    {	
	//kdWarning()<<" canvas_view::update -> number "<<i<<endl;
	if ( !m_data->isValidRect(i) )
	{
	    //kdWarning()<<"canvas_view::update  deleting "<<i<<endl;
	    if (m_rectlist[i])
		m_rectlist[i]->hide();
	    delete m_rectlist[i];
	    m_rectlist[i]=NULL;
	}
	else
	{
	    if (!m_rectlist[i])
	    {
		//kdWarning()<<"canvas_view::update create"<<i<<endl;
		m_rectlist[i] = new canvas_rect(canvas());
		m_rectlist[i]->show();
	    }

	    updateRect( i );
	    //m_rectlist[i]->checkFish();
	    //setRectColor(i);
	    //m_rectlist[i]->show();
	}
    }

    checkAllFish();

    canvas_rect::s_dofish=reset;

    canvas()->setAdvancePeriod(40);
    unlock();
}

void canvas_view::loadSettings()
{
    update();
    for (int i=0; i< m_data->nbRect(); i++)
    {
	updateRect(i);
    }
}

bool canvas_view::tryLock()
{
    if (m_updating)
	return false;

    m_updating=true;
    return true;
}

void canvas_view::unlock()
{
    m_updating=false;
}

#include "canvas_view.moc"


