/***************************************************************************
 *   Copyright (C) 2003, 2004 by Thomas Nagy                               *
 *   tnagy256@yahoo.fr                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <qdragobject.h>
#include <qpainter.h>
#include <qpaintdevicemetrics.h>
#include <qframe.h>
#include <qlayout.h>

#include <kdialogbase.h>
#include <kconfigdialog.h>
#include <kglobal.h>
#include <klocale.h>
#include <kiconloader.h>
#include <kdeversion.h>
#include <kmenubar.h>
#include <kstatusbar.h>
#include <kkeydialog.h>
#include <ktabwidget.h>
#include <kfiledialog.h>

#include <kmessagebox.h>
#include <kio/netaccess.h>
#include <ktempfile.h>

#include <kdebug.h>
#include <kstandarddirs.h>
#include <kurl.h>
#include <kurldrag.h>
#include <kurlrequesterdlg.h>

#include <kedittoolbar.h>

#include <kstdaccel.h>
#include <kaction.h>
#include <kstdaction.h>

//////////
#include "computationdata.h"
#include "packingdata.h"

#include "dataview.h"
#include "canvas_view.h"
#include "logview.h"

//////////
#include "settings.h"
#include "gui_prefs.h"
#include "algo_prefs.h"
#include "algo_chooser.h"
#include "boxmanager.h"

#include "algo/algobib.h"
#include "algo/algorun.h"
#include "linpacker.h"


linpacker::linpacker() : KMainWindow( 0, "linpacker" ), DCOPObject("sinterface")
{
    KTabWidget *tabwidget = new KTabWidget(this);

    // tell the KMainWindow that this is indeed the main widget
    setCentralWidget(tabwidget);

    m_data = packingdata::getData();
    m_dataview  = new dataview(tabwidget, m_data);
    m_canvasview = new canvas_view(tabwidget, m_data);
    m_logview = new logview(tabwidget, m_data);

    m_algorun = algorun::getRunner();

    tabwidget->addTab(m_dataview , i18n("Parameters") );
    tabwidget->addTab(m_canvasview, i18n("Packing representation") );
    tabwidget->addTab(m_logview, i18n("Logged information") );

    // accept dnd
    setAcceptDrops(true);

    // setup the actions
    setupActions();

    // and a status bar
    statusBar()->show();

    // apply the saved mainwindow settings, if any, and ask the mainwindow
    // to automatically save settings if changed: window size, toolbar
    // position, icon size, etc.
    setAutoSaveSettings();

    // allow the view to change the statusbar
    //connect(m_view, SIGNAL(signalChangeStatusbar(const QString&)),
    //      this,   SLOT(changeStatusbar(const QString&)));

    updateAllViews();

    connect(m_canvasview, SIGNAL(over(const QString &) ), 
	    this, SLOT(setStaticMessage(const QString&)) );
    connect(m_data, SIGNAL(rectangle_changed(int)),
	    m_canvasview, SLOT(updateRect(int)) );
}

linpacker::~linpacker()
{
    delete m_dataview;
    delete m_canvasview;
    delete m_logview;
    delete m_data;
}

void linpacker::updateAllViews()
{
    m_canvasview->update();
    m_dataview->update();
    m_logview->update();
}

void linpacker::setupActions()
{
    KStdAction::quit(this, SLOT(fileQuit()), actionCollection());
    KStdAction::open(this, SLOT(fileOpen()), actionCollection());
    KStdAction::save(this, SLOT(fileSave()), actionCollection());
    KStdAction::saveAs(this, SLOT(fileSaveAs()), actionCollection());
    KStdAction::preferences(this, SLOT(optionsPreferences()), actionCollection());


    m_openRecentAction = KStdAction::openRecent(this, SLOT(openURL(const KURL&)), actionCollection());

    m_openRecentAction->loadEntries( kapp->config(), "RecentFiles" );
    
    new KAction(
	    i18n("&Algorithm settings ..."),
	    0, this, SLOT(optionsAlgorithms()),
	    actionCollection(),
	    "algosettings");
    
    new KAction(
	    i18n("&Launch the resolution ..."),
	    SmallIcon("launch"), 0,
	    this, SLOT(launch()),
	    actionCollection(),
	    "launch");

    new KAction(i18n("&Modify the packing ..."),
	    0, this, SLOT(manageTable()),
	    actionCollection(),
	    "managetable");

    new KAction(i18n("&Export to ..."),
	    0, this, SLOT(exportData()),
	    actionCollection(),
	    "exportData");

    createGUI();
}

void linpacker::optionsPreferences()
{
    if (KConfigDialog::showDialog("settings"))
	return;

    KConfigDialog *dialog = new KConfigDialog(this, "settings",
	    Settings::self(), KDialogBase::Swallow );

    dialog->addPage(
           new gui_prefs(NULL),
           i18n("Appearance"),
	   "view_tree"
           /* locate("data", "linpacker/pics/view_tree.png") */
           ); 

    // this is broken (kdelibs documentation)
    /*QFrame *fr = dialog->plainPage();
    
    QWidget *w = new gui_prefs(fr);
    QGridLayout *layout = new QGridLayout(fr);
    layout->addWidget(w, 0, 0);*/
    
    dialog->setCaption( i18n("Preferences") );

    connect(dialog, SIGNAL(settingsChanged()), m_canvasview, SLOT(loadSettings()));
    connect(dialog, SIGNAL(settingsChanged()), m_dataview, SLOT(loadSettings()));
    //connect(dialog, SIGNAL(settingsChanged()), m_logview, SLOT(update()));

    dialog->show();
    dialog->enableLinkedHelp(false);
}

void linpacker::optionsAlgorithms()
{
    algobib::getBib()->makeTmpSettings();
 
    KDialogBase *dlg = new KDialogBase(this, 0, true, 
	    i18n("Algorithm settings"), KDialogBase::Ok|KDialogBase::Cancel);
    dlg->setMainWidget( new algo_prefs(dlg) );

    if ( dlg->exec() == QDialog::Accepted)
    {
	// TODO : this is a bit ugly (temporary settings .. uhm)
	// but as long as it works ...
	algobib::getBib()->loadTmpSettings();
    }
}

void linpacker::setTemporaryMessage(const QString& text)
{
    statusBar()->message(text, 2000);
}

void linpacker::setStaticMessage(const QString& text)
{
    statusBar()->message(text);
}

void linpacker::dragEnterEvent(QDragEnterEvent *event)
{
    // accept url drops only
    event->accept(KURLDrag::canDecode(event));
}

void linpacker::dropEvent(QDropEvent *event)
{
    KURL::List urls;

    // see if we can decode a URI.. if not, just ignore it
    if (KURLDrag::decode(event, urls) && !urls.isEmpty())
    {
	// okay, we have a URI.. process it
	const KURL &url = urls.first();

	// load in the file
	if (m_data->openFile(url))
	    updateAllViews();
    }
}

void linpacker::launch()
{
    //kdWarning()<<"launch"<<endl;

    if (m_data->m_computing)
	return;
    
    if (! m_data->pbvalide())
    {
	setTemporaryMessage(i18n("The packing is empty or cannot be solved") );
	return;
    }
    
    // create the dialog
    KDialogBase *dlg = new KDialogBase(this, 0, true,
            i18n("Solver parameters"), KDialogBase::Ok|KDialogBase::Cancel);
    algo_chooser *ouch = new algo_chooser(dlg);
    dlg->setMainWidget( ouch );

    if ( dlg->exec() == QDialog::Accepted)
    {
	m_data->m_adata = ouch->m_adata;
	
	m_algorun->start();

	m_dataview->update();
	m_canvasview->update();
	m_logview->update();
    } 
}

// TODO fix the dcop interface
void linpacker::start_solver()
{
    m_algorun->start();
}

void linpacker::manageTable()
{
    boxmanager * manager = new boxmanager(this, "boxmanager", m_data);
    manager->exec();
    m_dataview->update();
    m_canvasview->update();
}

void linpacker::exportData()
{
    KURL url = KFileDialog::getSaveURL(QString::null,
	    i18n("*.png|PNG picture (*.png)\n*.svg|SVG picture (*.svg)\n*.csv|Data file (*.csv)"),
	    this, i18n("Export to"));

    if (! url.isValid())
    {
	KMessageBox::sorry(this, i18n("The url or file selected is invalid") );
	return;
    }

    KTempFile docfile;
    docfile.setAutoDelete(true);

    if (url.url().endsWith("png") )
	m_canvasview->exportPNG(docfile.name());
    else if (url.url().endsWith("svg") )
	m_canvasview->exportSVG(docfile.textStream());
    else if (url.url().endsWith("csv") )
	m_data->exportCSV(docfile.textStream());

    docfile.close();

    if ( ! KIO::NetAccess::upload(docfile.name(), url, NULL) )
    {
	KMessageBox::sorry(this, i18n("The file could not be exported") );
    }
}

void linpacker::fileOpen()
{
    KURL url = KFileDialog::getOpenURL(QString::null, 
	    i18n("*.lnpk|linpacker files (*.lnpk)\n*.csv|CSV files (*.csv)"), 
	    this, i18n("Open Project"));

    if (url.isEmpty())
	return;
    
    openURL(url);
}

void linpacker::openURL(const KURL & url)
{
    m_data->openFile(url);
    m_url = url;
    m_openRecentAction->addURL(url);
    updateAllViews();
}

void linpacker::fileQuit()
{
    close();
}

void linpacker::fileSave()
{
    if (! m_data->saveFile(m_url))
	fileSaveAs();
}

void linpacker::fileSaveAs()
{
    KURL file_url = KFileDialog::getSaveURL(QString::null,
	    i18n("*.lnpk|linpacker file (*.lnpk)"),
	    this, i18n("Save Project"));
    if (file_url.isValid())
    {
	m_url = file_url;
	fileSave();
    }
}

bool linpacker::queryClose()
{
        // save the recent files
	m_openRecentAction->saveEntries( kapp->config(), "RecentFiles" );

        if (!m_data->m_ischanged)
        {
                //the document has not been changed, so we can just close
                return true;
        }

        // the document has changed, the user may want to save it
        int ans=KMessageBox::questionYesNoCancel(
                        0,
                        i18n("The current document has been changed. Do you want to save your changes?"),
                        i18n("Save changes?"),
                        i18n("Save"),
                        i18n("Discard"));

        switch (ans)
        {
                case KMessageBox::Cancel:
                        return false;
                        break;
                case KMessageBox::Yes:
                        fileSave();
			m_openRecentAction->saveEntries( kapp->config(), "RecentFiles" );
                        break;
                default:
                        break;
        }

	return true;
}

#include "linpacker.moc"

